<?php
declare(strict_types=1);
namespace Neos\EventSourcedContentRepository\Domain\Context\Workspace\Command;

/*
 * This file is part of the Neos.ContentRepository package.
 *
 * (c) Contributors of the Neos Project - www.neos.io
 *
 * This package is Open Source Software. For the full copyright and license
 * information, please view the LICENSE file which was distributed with this
 * source code.
 */

use Neos\ContentRepository\Domain\ContentStream\ContentStreamIdentifier;
use Neos\EventSourcedContentRepository\Domain\ValueObject\UserIdentifier;
use Neos\EventSourcedContentRepository\Domain\ValueObject\WorkspaceDescription;
use Neos\EventSourcedContentRepository\Domain\ValueObject\WorkspaceName;
use Neos\EventSourcedContentRepository\Domain\ValueObject\WorkspaceTitle;
use Neos\Flow\Annotations as Flow;

/**
 * Create a new workspace, based on an existing baseWorkspace
 *
 * @Flow\Proxy(false)
 */
final class CreateWorkspace
{
    private WorkspaceName $workspaceName;

    private WorkspaceName $baseWorkspaceName;

    private WorkspaceTitle $workspaceTitle;

    private WorkspaceDescription $workspaceDescription;

    private UserIdentifier $initiatingUserIdentifier;

    /**
     * the content stream identifier for the content stream which is created together with the to-be-created workspace
     */
    private ContentStreamIdentifier $newContentStreamIdentifier;

    private ?UserIdentifier $workspaceOwner;

    public function __construct(
        WorkspaceName $workspaceName,
        WorkspaceName $baseWorkspaceName,
        WorkspaceTitle $workspaceTitle,
        WorkspaceDescription $workspaceDescription,
        UserIdentifier $initiatingUserIdentifier,
        ContentStreamIdentifier $newContentStreamIdentifier = null,
        UserIdentifier $workspaceOwner = null
    ) {
        $this->workspaceName = $workspaceName;
        $this->baseWorkspaceName = $baseWorkspaceName;
        $this->workspaceTitle = $workspaceTitle;
        $this->workspaceDescription = $workspaceDescription;
        $this->initiatingUserIdentifier = $initiatingUserIdentifier;
        $this->newContentStreamIdentifier = $newContentStreamIdentifier ?: ContentStreamIdentifier::create();
        $this->workspaceOwner = $workspaceOwner;
    }

    public static function fromArray(array $array): self
    {
        return new static(
            new WorkspaceName($array['workspaceName']),
            new WorkspaceName($array['baseWorkspaceName']),
            new WorkspaceTitle($array['workspaceTitle']),
            new WorkspaceDescription($array['workspaceDescription']),
            UserIdentifier::fromString($array['initiatingUserIdentifier']),
            isset($array['newContentStreamIdentifier']) ? ContentStreamIdentifier::fromString($array['newContentStreamIdentifier']) : null,
            isset($array['workspaceOwner']) ? UserIdentifier::fromString($array['workspaceOwner']) : null
        );
    }

    public function getWorkspaceName(): WorkspaceName
    {
        return $this->workspaceName;
    }

    public function getBaseWorkspaceName()
    {
        return $this->baseWorkspaceName;
    }

    public function getWorkspaceTitle(): WorkspaceTitle
    {
        return $this->workspaceTitle;
    }

    public function getWorkspaceDescription(): WorkspaceDescription
    {
        return $this->workspaceDescription;
    }

    public function getInitiatingUserIdentifier(): UserIdentifier
    {
        return $this->initiatingUserIdentifier;
    }

    public function getNewContentStreamIdentifier(): ContentStreamIdentifier
    {
        return $this->newContentStreamIdentifier;
    }

    public function getWorkspaceOwner(): ?UserIdentifier
    {
        return $this->workspaceOwner;
    }
}
